import copy
import pygame
from pygame.locals import USEREVENT
import spaceship


class Pickup(pygame.sprite.Sprite):
    def __init__(self, action, **kwargs):   # image=None, image_file="", *args
        super().__init__()

        if kwargs.get('image', None) is not None:
            self._image = kwargs['image']
        else:
            self._image = pygame.image.load(kwargs['image_file'])

        self._rect = self._image.get_rect()
        self._action = action
        self._args = kwargs['args']

    def new_pickup(self):
        return Pickup(action=self._action, args=self._args, image=copy.copy(self._image))

    def move(self, vec):
        self._rect.center = (self._rect.center[0] + vec[0], self._rect.center[1] + vec[1])

    def render(self, window):
        window.blit(self._image, (self._rect.center[0] - self._rect.width//2,
                                  self._rect.center[1] - self._rect.height//2))

    @property
    def hit_box(self):
        return self._rect

    @property
    def image(self):
        return self._image

    @property
    def position(self):
        return self._rect.center

    @position.setter
    def position(self, vec):
        self._rect.center = vec

    @property
    def action(self):
        return self._action

    @property
    def args(self):
        return self._args

    @staticmethod
    def action_heal(entity, amount):
        entity.heal(amount)

    @staticmethod
    def action_give_gun(entity, gun):
        entity.add_ammo(gun.new_gun())

    @staticmethod
    def action_give_life(entity):
        entity.add_life()

    @staticmethod
    def action_fire_rate_boost(entity, multiplier, duration):
        entity.set_fire_rate_multiplier(multiplier)
        pygame.time.set_timer(RESET_FIRE_RATE, duration)

    @staticmethod
    def action_reset_fire_rate(entity):
        entity.set_fire_rate_multiplier(1)


RESET_FIRE_RATE = USEREVENT + 1

PICKUP_REPAIR = Pickup(
    action=Pickup.action_heal,
    image_file="Sprites/repair.png",
    args=[1000000]
)
PICKUP_SMALL_HEAL = Pickup(
    action=Pickup.action_heal,
    image_file="Sprites/small_heal.png",
    args=[5]
)
PICKUP_MEDIUM_HEAL = Pickup(
    action=Pickup.action_heal,
    image_file="Sprites/medium_heal.png",
    args=[10]
)
PICKUP_MINIGUN = Pickup(
    action=Pickup.action_give_gun,
    image_file="Sprites/minigun_pickup.png",
    args=[spaceship.MINIGUN]
)
PICKUP_ROCKET_GUN = Pickup(
    action=Pickup.action_give_gun,
    image_file="Sprites/rocket_gun_pickup.png",
    args=[spaceship.ROCKET_GUN]
)
PICKUP_LIFE = Pickup(
    action=Pickup.action_give_life,
    image_file="Sprites/life.png",
    args=[]
)
PICKUP_FIRE_RATE_BOOSTER = Pickup(
    action=Pickup.action_fire_rate_boost,
    image_file="Sprites/fire_rate_booster.png",
    args=[0.5, 5000]
)
